#include "Graph.hpp"
#include "HtorSets.hpp"
#include "DataFlow.hpp"
#include "config.h"
#include <set>
#include <vector>
// some trickery to get uint64_t --- you would think the natural
// way to do this would be to check for csdtint in the configure
// script and then just include it, however, cstdint only works
// with -std=c++0x at the moment --- this would crush all hopes
// for portable code

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif

#include <stdio.h>
#include <algorithm>

#define T_SEND 1
#define T_RECV 2
#define T_DUMMY 3



bool TupleStartSorter (const tuple* a, const tuple* b) {
    return a->start < b->start;
}

void split(std::vector<tuple*> *Or_ref, tuple *s, std::set<tuple*> *Is_ref, std::set<tuple*> *S_ref, std::set<tuple*> *R_ref, std::set<tuple*> *N_ref, std::map<tuple*, tuple*> *M_ref, std::vector< /* procs */ std::vector< std::pair<tuple*,tuple*> > > *D_ref) {

  std::vector<tuple*>& Or= *Or_ref; 
  std::set<tuple*>& S= *S_ref; 
  std::set<tuple*>& R= *R_ref; 
  std::set<tuple*>& N= *N_ref; 
  std::set<tuple*>& Ss= *Is_ref; 
  std::map<tuple*, tuple*>& M=*M_ref;
  std::vector< /* procs */ std::vector< std::pair<tuple*,tuple*> > >& D= *D_ref;

  if(Or.size() == 1 && Or[0]->size == s->size && Or[0]->start == s->start) {
    s->peer = Or[0]->peer;
    s->sowner = Or[0]->sowner;
    s->sstart = Or[0]->sstart;
    s->ssize = Or[0]->ssize;
    s->smtype = Or[0]->smtype;
    Ss.insert(s);
    //printf("split succeeded -- %i\n", s->id);
  }
   else {
    // order Or by start address
    std::sort(Or.begin(), Or.end(), TupleStartSorter);
    // find r such that (s,r) \in M
    //tuple *r = M[s];
    //printf("in split id: %i\n", s->id);
    std::map<tuple*, tuple*>::iterator titer = M.find(s);
    assert(titer != M.end());
    tuple *r = (*titer).second;
    
    uint64_t riter = r->start;
    tuple *dr = new tuple();
    dr->type = T_DUMMY;
    dr->owner = r->owner;
    dr->id= -101;
    N.insert(dr);
    tuple *ds = new tuple();
    ds->type = T_DUMMY;
    ds->owner = s->owner;
    ds->id= -100;
    N.insert(ds);

    // remove s from S and r from R
    std::set<tuple*>::iterator rit = R.find(r);
    assert(rit != R.end());
    R.erase(rit);
    std::set<tuple*>::iterator sit = S.find(s);
    assert(sit != S.end());
    S.erase(sit);

    // replace r with dr
    for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[r->owner].begin(); x!=D[r->owner].end(); x++) {
      if(r == x->first) x->first = dr;
      if(r == x->second) x->second = dr;
    }
    // replace s with ds
    for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[s->owner].begin(); x!=D[s->owner].end(); x++) {
      if(s == x->first) x->first = ds;
      if(s == x->second) x->second = ds;
    }

    // remove (s,r) from M
    std::map<tuple*, tuple*>::iterator mit = M.find(s);
    assert(mit != M.end());
    M.erase(mit);
    
    uint64_t siter = s->start;
    // foreach r \in Or (ordered)
    for(std::vector<tuple*>::iterator rit=Or.begin(); rit!=Or.end(); ++rit) {
      tuple *ir = *rit;
      //printf("recv: %i-%i (type: %i)\n", r->start, r->start+r->size, r->mtype);
      uint64_t locrange = std::max(ir->start, siter)-siter;
      if(locrange > 0) {
        //printf("split locrange: %i\n", locrange);
        tuple *ns = new tuple();
        ns->type = T_SEND;
        ns->id = -1*s->id;
        ns->owner = s->owner;
        ns->start = siter;
        ns->size = locrange;
        ns->peer = s->peer;
        ns->mtype = s->mtype;

        ns->sowner = s->owner;
        ns->sstart = siter;
        ns->ssize = locrange;
        ns->smtype = s->mtype;

        tuple *nr = new tuple();
        nr->type = T_RECV;
        nr->id = -1*ir->id;
        nr->owner = r->owner;
        nr->start = riter;
        nr->size = locrange;
        nr->peer = ir->peer;
        nr->mtype = ir->mtype;

        nr->sowner = s->sowner;
        nr->sstart = siter;
        nr->ssize = locrange;
        nr->smtype = s->smtype; 
        
        S.insert(ns);
        Ss.insert(ns);
        R.insert(nr);

        D[r->owner].push_back(std::make_pair(dr, nr));
        D[s->owner].push_back(std::make_pair(ns, ds));
        M[ns] = nr ;
        siter += locrange;
        riter += locrange;
      }
      int remrange = std::min(ir->start+ir->size, s->start+s->size)-siter;
      //printf("split remrange: %i\n", remrange);

      tuple *ns = new tuple();
      ns->type = T_SEND;
      ns->id = -1*s->id;
      ns->owner = s->owner;
      ns->start = siter;
      ns->size = remrange;
      ns->peer = s->peer;
      ns->mtype = s->mtype;

      ns->sowner = ir->sowner;
      ns->sstart = ir->sstart;
      ns->ssize = remrange;
      ns->smtype = ir->smtype;

      tuple *nr = new tuple();
      nr->type = T_RECV;
      nr->id = -1*r->id;
      nr->owner = r->owner;
      nr->start = riter;
      nr->size = remrange;
      nr->peer = r->peer;
      nr->mtype = r->mtype;

      nr->sowner = ir->sowner;
      nr->sstart = ir->sstart;
      nr->ssize = remrange;
      nr->smtype = ir->smtype; 
      
      S.insert(ns);
      Ss.insert(ns);
      R.insert(nr);

      D[r->owner].push_back(std::make_pair(dr, nr));
      D[s->owner].push_back(std::make_pair(ns, ds));
      M[ns] =nr;
      siter += remrange;
      riter += remrange;
	}
	delete(s);
	delete(r);

    //printf("split failed, size: %i!\n", Or.size());
  }
}


std::vector<tuple*>* DataFlowAnalysis(HtorSets* hs) {

  int p=0;
  std::vector<tuple*>* tuples = new std::vector<tuple*>;
  
  std::map<tuple*, tuple*> M;
  std::map<uint64_t, tuple*> tmpmaps;
  std::map<uint64_t, tuple*> tmpmapr;
  std::vector< /* procs */ std::vector< std::pair<tuple*,tuple*> > > D;
  std::set<tuple*> S;
  std::set<tuple*> R;
  std::set<tuple*> N; // set of all no-ops

  // initialize all r \in R with (owner:start:size:NIL:NIL:NIL);
  for(std::vector<Node*>::iterator it = hs->Recvs.begin(); it!=hs->Recvs.end(); ++it) {
    tuple *t = new tuple();
    t->type = T_RECV;
    t->id = (*it)->getKey();
    t->owner = (*it)->getOwner();
    t->start = (uint64_t) (*it)->getAddr();
    t->mtype = (*it)->getMemtype();
    t->size = (*it)->getSize();
    t->peer = (*it)->getPeer();
    t->sstart = 0xffffffff;
    t->ssize = -1;
    t->smtype = -1;
    t->sowner = -1;
    R.insert(t);
	tmpmapr[t->id] = t;
    
	//printf("insert recv key %i\n", t->id);
    // find p -- TODO: replace ;)
    if(p<t->owner) p=t->owner;
  }
  
  // initialize all s \in S with (owner:start:size);
  for(std::vector<Node*>::iterator it = hs->Sends.begin(); it!=hs->Sends.end(); ++it) {
    tuple *t = new tuple();
    t->type = T_SEND;
    t->id = (*it)->getKey();
    t->owner = (*it)->getOwner();
    t->start = (uint64_t) (*it)->getAddr();
    t->mtype = (*it)->getMemtype();
    t->size = (*it)->getSize();
    t->sstart = 0xffffffff;
    t->ssize = -1;
    t->smtype = -1;
    t->sowner = -1;
	S.insert(t);
    tmpmaps[t->id] = t;
    //printf("insert send key %i\n", t->id);
    // find p -- TODO: replace ;)
    if(p<t->owner) p=t->owner;
  }
  p++; // we start numbering procs at 0 -- TODO: really get rid of this
  
  // insert matches into map (set) M using tmpmap
  // operations on M: look up matching recv for a send
  for(std::vector< std::pair<Node*,Node*> >::iterator it = hs->Matches.begin(); it!=hs->Matches.end(); ++it) {
    tuple *s = tmpmaps[it->first->getKey()];
    tuple *r = tmpmapr[it->second->getKey()];
    M[s] = r;
  }
  
  // insert all dependency pairs into map (for lookup later)
  // operations on D: look up all (x,s) that are not in D *and* remove all (r,x)
  // a simple way to implement this is a traversal (but slow),
  // dependencies can only be inside one node, thus, we have a vector of
  // pairs for each node to limit traversal overhead 
  // TODO: enhance further with double-linked strategy!
  D.resize(p);
  for(std::vector< std::pair<Node*,Node*> >::iterator it = hs->Dependencies.begin(); it!=hs->Dependencies.end(); ++it) {
    //tuple *s = tmpmaps[it->first->getKey()];
    //tuple *r = tmpmapr[it->second->getKey()];
    //printf("trying to find send key %i\n", it->first->getKey());
    std::map<uint64_t, tuple*>::iterator siter = tmpmaps.find(it->first->getKey());
    assert(siter != tmpmaps.end());
    tuple *s = (*siter).second;
    //printf("trying to find recv key %i\n", it->second->getKey());
    std::map<uint64_t, tuple*>::iterator riter = tmpmapr.find(it->second->getKey());
    assert(riter != tmpmapr.end());
    tuple *r = (*riter).second;

    assert(s->owner == r->owner);
    D[s->owner].push_back(std::make_pair(r,s));
  }

  // finished receives 
  std::vector< std::set<tuple*> > Fr(p);
  // finished sends
  std::set<tuple*> Fs;
  // independent sends
  std::set<tuple*> Is;

  // initialization of data structures is finished here
 
  // delete unneeded stuff
  tmpmaps.clear();
  tmpmapr.clear();


 
  
  // find set Is \subset S of independent sends, such that s \in Is -> (x,s) \notin D
  for (std::set<tuple*>::iterator sit = S.begin(); sit!=S.end(); ++sit) {
    tuple *s=*sit;
    int r=s->owner;
    // check if (x,s) \in D
    int found=0;
    for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[r].begin(); x!=D[r].end(); ++x) {
      if(s == x->second) { found = 1; break; }
    }
    // add s to Is
    if(!found) Is.insert(s);
  }
  
  // while Is \not \emptyset
  while(! Is.empty() ) {
    // foreach s \in Is
    for (std::set<tuple*>::iterator sit = Is.begin(); sit!=Is.end(); ++sit) {
      // find r \in R such that (s,r) \in M
      tuple *s = *sit;
      tuple *r = M[s];
	  assert(s != 0);
	  assert(r != 0);
      // update r tuple to X:X:X:s.owner:s.start:s.size
      assert(r->peer == s->owner);
      if(s->sstart != 0xffffffff) { 
        //r->peer = s->peer;
        r->sowner = s->sowner;
        r->sstart = s->sstart;
        r->ssize = s->ssize;
        r->smtype = s->smtype;
      } else {
        r->sowner = s->owner;
        r->sstart = s->start;
        r->ssize = s->size;
        r->smtype = s->mtype;
      }
      // remove (r,*) from D -- TODO: this is horribly slow due to the vector stuff :-(
      for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[r->owner].begin(); x!=D[r->owner].end(); ) {
        if(r == x->first) x = D[r->owner].erase(x);
        else x++;
      }
      
      // r -> Fr
      Fr[r->owner].insert(r);
      //printf("finish receive on %i from %i (%i) - %i %i %i\n", r->owner, r->src, r->id, r->sstart, r->ssize, r->smtype);
      // s -> Fs
      Fs.insert(s);
      //printf("finish send on %i (%i)\n", s->owner, s->id);
    }
    
    // remove all independent no-ops (n,*) from D -- TODO: this is horribly slow due to the vector stuff :-(
    // -- first, find all independent no-ops
    std::set<tuple*> In; // independent no-ops
    for (std::set<tuple*>::iterator nit = N.begin(); nit!=N.end(); ++nit) {
      tuple *n=*nit;
      int r=n->owner;
      // check if (x,s) \in D TODO: factor out as function (is same as above)
      int found=0;
      for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[r].begin(); x!=D[r].end(); ++x) {
        if(n == x->second) { found = 1; break; }
      }
      if(!found) In.insert(n);
      //printf("found %i independent no-ops\n", In.size());
    }
    // -- now, remove them all
    for (std::set<tuple*>::iterator nit = In.begin(); nit!=In.end(); ++nit) {
      tuple *n=*nit;
      int found=0;
      for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[n->owner].begin(); x!=D[n->owner].end(); ) {
        if(n == x->first) {x = D[n->owner].erase(x); found=1; } else x++;
      }
      // remove n from N
      if(found) {
        std::set<tuple*>::iterator nit = N.find(n);
        assert(nit != N.end());
		delete(n);
        N.erase(nit);
        //printf("removed no-op id: %i, %i left\n", n->id, N.size());
        // TODO: this is full of memory leaks -- need to destroy ops, not for now ;)
      }
	  else {
        printf("not found in D??? (should never happen)\n");
      }
    }

    // find As such that s \in As -> (x,s) \notin D \and s \notin Fs
    std::set<tuple*> As;
    // find set Is \subset S of independent sends, such that s \in Is -> (x,s) \notin D
    for (std::set<tuple*>::iterator sit = S.begin(); sit!=S.end(); ++sit) {
      tuple *s=*sit;
      int r=s->owner;
      // check if (x,s) \in D TODO: factor out as function (is same as above)
      int found=0;
      for(std::vector< std::pair<tuple*,tuple*> >::iterator x = D[r].begin(); x!=D[r].end(); ++x) {
        if(s == x->second) { found = 1; break; }
      }
      if(!found && Fs.find(s) != Fs.end()) found = 1;
      //if(!found) printf("found active send on %i\n", s->owner);
      // add s to As
      if(!found) As.insert(s);
    }
    // Is = \emptyset
    Is.clear();
    // foreach a \in As 
    for (std::set<tuple*>::iterator ai = As.begin(); ai!=As.end(); ++ai) {
      std::vector<tuple*> Or;
      // find overlapping receive -- Fr is a vector of p sets -- purely
      // for performance reasons!
      tuple *a = *ai;
      for(std::set<tuple*>::iterator ri = Fr[a->owner].begin(); ri != Fr[a->owner].end(); ++ri) {
        tuple *r = *ri;
        //printf("%i (%i) %i\n", r->owner, r->id, a->owner);
        //if(r->owner == a->owner) {
          uint64_t rstart = r->start;
          uint64_t rend = rstart + r->size;
          uint64_t astart = a->start;
          uint64_t aend = astart + a->size;
          //printf("%i<-%i %i-%i (in region %i-%i)\n", r->owner, r->peer, rstart, rend, astart, aend);
          if(! (rstart >= aend || rend <= astart) ) { // overlap
            Or.push_back(r);
          }
        //}
      }
      if(!Or.empty()) split(&Or, a, &Is, &S, &R, &N, &M, &D);
    }


  }
  // if(|Fr| != |R| \or |Fs| != |S|) return unmatched Send/Recv or cycle

// timos: keeping all the original tupels is bad, memory-wise, so remove this check
//  int frsize = 0;
//  for(int i=0; i<Fr.size(); ++i) frsize += Fr[i].size();
//  if(frsize != R.size() || Fs.size() != S.size()) printf("unmatched Send/Recv or cycle (%i!=%i or %i!=%i)\n", Fr.size(), R.size(), Fs.size(), S.size());
//  printf("don't forget to merge consecutive tuples with the same src!\n");

	for (std::set<tuple*>::iterator sit = S.begin(); sit!=S.end(); ++sit) {
		delete(*sit);
	}	

	for (std::set<tuple*>::iterator nit = N.begin(); nit!=N.end(); ++nit) {
		delete(*nit);
	}

	for(int i=0; i<Fr.size(); ++i) {
		for (std::set<tuple*>::iterator ri = Fr[i].begin(); ri!=Fr[i].end(); ++ri) {
			tuple *t = *ri;
#ifndef NDEBUG
			if(t->mtype == MEM_SCRATCH) printf("rcvd into scratch - check: ");
			if(t->smtype == MEM_SCRATCH) printf("sent from scratch - check: ");
#endif
    		if(t->mtype == MEM_SCRATCH || t->smtype == MEM_SCRATCH) {
#ifndef NDEBUG
 				printf("id: %i, owner: %i, range: (%x,%i,%i), from: %i, range: (%x,%i,%i)\n", t->id, t->owner, t->start, t->size, (int)t->mtype, t->sowner, t->sstart, t->ssize, (int)t->smtype);
#endif
 			}
			else {
				tuples->push_back(t);
			}
		}
 	}

	return tuples;

}

