#include <mpi.h>
#include "tests.hpp"
#include "libcDAG.hpp"

void init_buffer(void* buffer, size_t buffer_size) {
	
	srand(time(NULL));
	
	for (size_t i=0; i<buffer_size; i++) {
		char rnd = rand() % 256;
		((char*) buffer)[i] = rnd;
	}

}

int* figure_out_random_peers() {

	const float msg_xtra = 1.0;

	int p, r;
	MPI_Comm_size(MPI_COMM_WORLD, &p);
	MPI_Comm_rank(MPI_COMM_WORLD, &r);

	/*
	We want msg_xtra*p extra messages in the network, between random nodes
	of course each send needs a matching recv, so figure out who will
	communicate with who and tell the peers about that
	*/

	int* msgs = (int*) calloc(p*p,sizeof(int));

	if (r == 0) {
		srand(time(NULL));
		for (int i=0; i<p; i++) {
			int sender = 0;
			int receiver = 0;
			while (sender == receiver) {
				sender = rand() % p;
				receiver = rand() % p;
			}
			msgs[receiver*p+sender]++;
		}	
	}

	// That's a waste of bandwidth to bcast the whole array, but it's not time critical anyway
	MPI_Bcast(msgs, p*p, MPI_INT, 0, MPI_COMM_WORLD);

	return msgs;
		
}

int main(int argc, char** argv) {

	MPI_Init(&argc, &argv);
	TEST_INIT("linear broadcast");

	cDAG_Init();

	const int datasize = 8;
	
	int r = cDAG_MyRank();
	int p = cDAG_GroupSize();
	char* buffer_cdag   = (char*) malloc(datasize);
	char* buffer_mpi    = (char*) malloc(datasize);
	char* random_buffer = (char*) malloc(2*datasize);
	init_buffer(buffer_cdag, datasize);
	init_buffer(buffer_mpi, datasize);
	init_buffer(random_buffer, 2*datasize);

	int* msgs = figure_out_random_peers();

for (int iter=0; iter<1000; iter++) {
	cDAG_Graph g = cDAG_Create_graph();

	/* the linear bcast is built here */
	if (r == 0) {
		for (int i=1; i<p; i++) {
			cDAG_Send(g, buffer_cdag, datasize, i);
		}
	}
	else {
		cDAG_Recv(g, buffer_cdag, datasize, 0);
	}

	/* the random msgs are added now (sends)*/
	for (int peer=0; peer<p; peer++) {
		int n_sends = msgs[peer*p+r];
		for (int m=0; m<n_sends; m++) { // note that n_sends should be either 0 (most cases) or 1, however the possibility of it beeing > 1 is not null
			cDAG_Send(g, random_buffer, datasize*2, peer);
		}
		int n_recvs = msgs[r*p+peer];
		for (int m=0; m<n_recvs; m++) {
			cDAG_Recv(g, random_buffer, datasize*2, peer);
		}
	}

	g->do_coll_detection = 0;
	cDAG_Schedule sched = cDAG_Compile(g);
	cDAG_FreeGraph(g);
	cDAG_Handle hndl = cDAG_Run(sched);
	cDAG_Wait(hndl);

	cDAG_FreeSchedule(sched);

}

	MPI_Bcast(buffer_mpi, datasize, MPI_BYTE, 0, MPI_COMM_WORLD);

	if (memcmp(buffer_mpi, buffer_cdag, datasize) != 0) TEST_FAILED();
	else TEST_SUCCESSFULL();

	cDAG_Finalize();
	MPI_Finalize();

	free(buffer_cdag);
	free(buffer_mpi);
	free(random_buffer);
	free(msgs);

}
